﻿// Copyright (c) Microsoft Corporation. All rights reserved.
// Licensed under the MIT License.

using System;
using Microsoft.Identity.Client;
using Microsoft.Identity.Client.Internal.Requests;
using Microsoft.Identity.Test.Common;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Microsoft.Identity.Client.Internal;
using System.Security.Cryptography.X509Certificates;
using System.Collections.Generic;

namespace Microsoft.Identity.Test.Unit.RequestsTests
{
    [TestClass]
    [DeploymentItem(@"Resources\testCert.crtfile")]
    public class RequestValidationHelperTests
    {
        public const uint JwtToAadLifetimeInSeconds = 60 * 10; // Ten minutes

        private IServiceBundle _serviceBundle;
        readonly string _audience1 = "Audience1";
        readonly string _audience2 = "Audience2";

        [TestInitialize]
        public void TestInitialize()
        {
            TestCommon.ResetInternalStaticCaches();
            _serviceBundle = TestCommon.CreateDefaultServiceBundle();
        }

        [TestMethod]
        [Description("Test for client assertion with mismatched parameters in Request Validator.")]
        public void ClientAssertionRequestValidatorMismatchParameterTest()
        {
            var credential = ClientCredentialWrapper.CreateWithSecret(TestConstants.ClientSecret);
            credential.Audience = _audience1;
            credential.ContainsX5C = false;
            credential.CachedAssertion = TestConstants.DefaultClientAssertion;
            credential.ValidTo = ConvertToTimeT(DateTime.UtcNow + TimeSpan.FromSeconds(JwtToAadLifetimeInSeconds));

            // Validate cached client assertion with parameters
            Assert.IsTrue(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));

            // Different audience
            credential.Audience = _audience2;

            // cached assertion should be invalid
            Assert.IsFalse(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));

            // Different x5c, same audience
            credential.Audience = _audience1;
            credential.ContainsX5C = true;

            // cached assertion should be invalid
            Assert.IsFalse(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));

            // Different audience and x5c
            credential.Audience = _audience2;

            // cached assertion should be invalid
            Assert.IsFalse(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));

            // No cached Assertion
            credential.CachedAssertion = "";

            // should return false
            Assert.IsFalse(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));
        }

        [TestMethod]
        [Description("Test for expired client assertion in Request Validator.")]
        public void ClientAssertionRequestValidatorExpirationTimeTest()
        {
            var credential = ClientCredentialWrapper.CreateWithSecret(TestConstants.ClientSecret);
            credential.Audience = _audience1;
            credential.ContainsX5C = false;
            credential.CachedAssertion = TestConstants.DefaultClientAssertion;
            credential.ValidTo = ConvertToTimeT(DateTime.UtcNow + TimeSpan.FromSeconds(JwtToAadLifetimeInSeconds));

            // Validate cached client assertion with expiration time
            // Cached assertion should be valid
            Assert.IsTrue(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));

            // Setting expiration time to now
            credential.ValidTo = ConvertToTimeT(DateTime.UtcNow);

            // cached assertion should have expired
            Assert.IsFalse(ClientCredentialHelper.ValidateClientAssertion(credential, _audience1, false));
        }

        [TestMethod]
        public void CCACreatedWithoutAuthenticationType_Throws()
        {
            ApplicationConfiguration config = new ApplicationConfiguration();

           Assert.ThrowsException<MsalClientException>(
               () => new ClientCredentialWrapper(config));
        }

        [TestMethod]
        public void CCACreatedWithAuthenticationType_ClientSecret_DoesNotThrow()
        {
            // Arrange
            ApplicationConfiguration config = new ApplicationConfiguration
            {
                ClientSecret = TestConstants.ClientSecret
            };

            // Act
            ClientCredentialWrapper clientCredentialWrapper = new ClientCredentialWrapper(config);

            // Assert
            // no exception is thrown
            Assert.AreEqual(
                ConfidentialClientAuthenticationType.ClientSecret, 
                clientCredentialWrapper.AuthenticationType);
        }

        [TestMethod]
        public void CCACreatedWithAuthenticationType_ClientCertificate_DoesNotThrow()
        {
            // Arrange
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");
            ApplicationConfiguration config = new ApplicationConfiguration
            {
                ClientCredentialCertificate = cert
            };

            // Act
            ClientCredentialWrapper clientCredentialWrapper = new ClientCredentialWrapper(config);

            // Assert
            // no exception is thrown
            Assert.AreEqual(
                ConfidentialClientAuthenticationType.ClientCertificate,
                clientCredentialWrapper.AuthenticationType);
        }

        [TestMethod]
        public void CCACreatedWithAuthenticationType_SignedClientAssertion_DoesNotThrow()
        {
            // Arrange
            ApplicationConfiguration config = new ApplicationConfiguration
            {
                SignedClientAssertion = "signed"
            };

            // Act
            ClientCredentialWrapper clientCredentialWrapper = new ClientCredentialWrapper(config);

            // Assert
            // no exception is thrown
            Assert.AreEqual(
                ConfidentialClientAuthenticationType.SignedClientAssertion,
                clientCredentialWrapper.AuthenticationType);
        }

        [TestMethod]
        public void CCACreatedWithAuthenticationType_ClientCertificateWithNoClaims_DoesNotThrow()
        {
            // Arrange
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");
            var claims = new Dictionary<string, string>();
            ApplicationConfiguration config = new ApplicationConfiguration
            {                
                ClientCredentialCertificate = cert,
                ClaimsToSign = claims
            };

            // Act
            ClientCredentialWrapper clientCredentialWrapper = new ClientCredentialWrapper(config);

            // Assert
            // no exception is thrown
            Assert.AreEqual(
                ConfidentialClientAuthenticationType.ClientCertificate,
                clientCredentialWrapper.AuthenticationType);
        }

        [TestMethod]
        public void CCACreatedWithAuthenticationType_ClientCertificateWithClaims_DoesNotThrow()
        {
            // Arrange
            var cert = new X509Certificate2(
                ResourceHelper.GetTestResourceRelativePath("testCert.crtfile"), "passw0rd!");
            var claims = new Dictionary<string, string>();
            claims.Add("cats", "are cool");

            ApplicationConfiguration config = new ApplicationConfiguration
            {
                ClientCredentialCertificate = cert,
                ClaimsToSign = claims
            };

            // Act
            ClientCredentialWrapper clientCredentialWrapper = new ClientCredentialWrapper(config);

            // Assert
            // no exception is thrown
            Assert.AreEqual(
                ConfidentialClientAuthenticationType.ClientCertificateWithClaims,
                clientCredentialWrapper.AuthenticationType);
        }

        internal static long ConvertToTimeT(DateTime time)
        {
            var startTime = new DateTime(1970, 1, 1, 0, 0, 0, 0);
            TimeSpan diff = time - startTime;
            return (long)diff.TotalSeconds;
        }
    }
}
